@@ -34,6 +34,7 @@
 
 #define TRNG_CHI_SQUARE_DIST_HPP
 
+#include <trng/cuda.hpp>
 #include <trng/limits.hpp>
 #include <trng/utility.hpp>
 #include <trng/math.hpp>
@@ -56,10 +57,14 @@ namespace trng {
     private:
       int nu_;
     public:
+      TRNG_CUDA_ENABLE
       int nu() const { return nu_; }
+      TRNG_CUDA_ENABLE
       void nu(int nu_new) { nu_=nu_new; }
+      TRNG_CUDA_ENABLE
       param_type() : nu_(1) {
       }
+      TRNG_CUDA_ENABLE
       explicit param_type(int nu) : nu_(nu) {
       }
 
@@ -102,6 +107,7 @@ namespace trng {
     param_type p;
 
     // inverse cumulative density function
+    TRNG_CUDA_ENABLE
     result_type icdf_(result_type x) const {
       if (x<=math::numeric_limits<result_type>::epsilon())
         return 0;
@@ -128,30 +134,42 @@ namespace trng {
     
   public:
     // constructor
+    TRNG_CUDA_ENABLE
     explicit chi_square_dist(int nu) : p(nu) {
     }
+    TRNG_CUDA_ENABLE
     explicit chi_square_dist(const param_type &p) : p(p) {
     }
     // reset internal state
+    TRNG_CUDA_ENABLE
     void reset() { }
     // random numbers
     template<typename R>
+    TRNG_CUDA_ENABLE
     result_type operator()(R &r) {
       return icdf_(utility::uniformco<result_type>(r));
     }
     template<typename R>
+    TRNG_CUDA_ENABLE
     result_type operator()(R &r, const param_type &p) {
       chi_square_dist g(p);
       return g(r);
     }
     // property methods
+    TRNG_CUDA_ENABLE
     result_type min() const { return 0; }
+    TRNG_CUDA_ENABLE
     result_type max() const { return math::numeric_limits<result_type>::infinity(); }
+    TRNG_CUDA_ENABLE
     param_type param() const { return p; }
+    TRNG_CUDA_ENABLE
     void param(const param_type &p_new) { p=p_new; }
+    TRNG_CUDA_ENABLE
     int nu() const { return p.nu(); }
+    TRNG_CUDA_ENABLE
     void nu(int nu_new) { p.nu(nu_new); }
     // probability density function  
+    TRNG_CUDA_ENABLE
     result_type pdf(result_type x) const {
       if (x<0)
 	return 0;
@@ -162,6 +180,7 @@ namespace trng {
       }
     }
     // cumulative density function 
+    TRNG_CUDA_ENABLE
     result_type cdf(result_type x) const {
       if (x<=0)
 	return 0;
@@ -169,9 +188,12 @@ namespace trng {
 	return math::GammaP(p.nu()/result_type(2), x/2);
     }
     // inverse cumulative density function 
+    TRNG_CUDA_ENABLE
     result_type icdf(result_type x) const {
       if (x<=0 or x>=1) {
+#if !(defined __CUDA_ARCH__)
 	errno=EDOM;
+#endif
 	return math::numeric_limits<result_type>::quiet_NaN();
       }
       if (x==0)
@@ -186,12 +208,14 @@ namespace trng {
 
   // EqualityComparable concept
   template<typename float_t>
+  TRNG_CUDA_ENABLE
   inline bool operator==(const typename chi_square_dist<float_t>::param_type &p1, 
 			 const typename chi_square_dist<float_t>::param_type &p2) {
     return p1.nu()==p2.nu();
   }
 
   template<typename float_t>
+  TRNG_CUDA_ENABLE
   inline bool operator!=(const typename chi_square_dist<float_t>::param_type &p1, 
 			 const typename chi_square_dist<float_t>::param_type &p2) {
     return not (p1==p2);
@@ -201,12 +225,14 @@ namespace trng {
 
   // EqualityComparable concept
   template<typename float_t>
+  TRNG_CUDA_ENABLE
   inline bool operator==(const chi_square_dist<float_t> &g1, 
 			 const chi_square_dist<float_t> &g2) {
     return g1.param()==g2.param();
   }
 
   template<typename float_t>
+  TRNG_CUDA_ENABLE
   inline bool operator!=(const chi_square_dist<float_t> &g1, 
 			 const chi_square_dist<float_t> &g2) {
     return g1.param()!=g2.param();
