@@ -34,6 +34,7 @@
 
 #define TRNG_LCG64_HPP
 
+#include <trng/cuda.hpp>
 #include <trng/limits.hpp>
 #include <climits>
 #include <stdexcept>
@@ -51,17 +52,22 @@ namespace trng {
 
     // Uniform random number generator concept
     typedef unsigned long long result_type;
+    TRNG_CUDA_ENABLE
     result_type operator()() const; 
     static const result_type min=0;
     static const result_type max=18446744073709551615ull;
   private:
     // compute floor(log_2(x))
+    TRNG_CUDA_ENABLE
     static unsigned int log2_floor(result_type x);
     // compute pow(x, n)
+    TRNG_CUDA_ENABLE
     static result_type pow(result_type x, result_type n);
     // compute prod(1+a^(2^i), i=0..l-1)
+    TRNG_CUDA_ENABLE
     static result_type g(unsigned int l, result_type a);
     // compute sum(a^i, i=0..s-1)
+    TRNG_CUDA_ENABLE
     static result_type f(result_type s, result_type a);
   public:
     // Parameter and status classes
@@ -226,12 +232,16 @@ namespace trng {
     }
     
     // Parallel random number generator concept
+    TRNG_CUDA_ENABLE
     void split(unsigned int, unsigned int);
+    TRNG_CUDA_ENABLE
     void jump2(unsigned int);
+    TRNG_CUDA_ENABLE
     void jump(unsigned long long);
     
     // Other useful methods
     static const char * name();
+    TRNG_CUDA_ENABLE
     long operator()(long) const;
 
   private:
@@ -239,12 +249,15 @@ namespace trng {
     mutable status_type S;
     static const char * const name_str;
     
+    TRNG_CUDA_ENABLE
     void backward();
+    TRNG_CUDA_ENABLE
     void step() const;
   };
   
   // Inline and template methods
   
+  TRNG_CUDA_ENABLE
   inline void lcg64::step() const {
 #if ULONG_LONG_MAX>18446744073709551615ull
     S.r=(P.a*S.r+P.b) & max;
@@ -253,15 +266,128 @@ namespace trng {
 #endif
   }
     
+  TRNG_CUDA_ENABLE
   inline lcg64::result_type lcg64::operator()() const {
     step();
     return S.r;
   }
 
+  TRNG_CUDA_ENABLE
   inline long lcg64::operator()(long x) const {
     return static_cast<long>(utility::uniformco<double, lcg64>(*this)*x);
   }
 
+  // compute floor(log_2(x))
+  TRNG_CUDA_ENABLE
+  inline unsigned int lcg64::log2_floor(lcg64::result_type x) {
+    unsigned int y(0);
+    while (x>0) {
+      x>>=1;
+      ++y;
+    };
+    --y;
+    return y;
+  }
+
+  // compute pow(x, n)
+  TRNG_CUDA_ENABLE
+  inline lcg64::result_type lcg64::pow(lcg64::result_type x, lcg64::result_type n) {
+    lcg64::result_type result=1;
+    while (n>0) {
+      if ((n&1)>0)
+	result=result*x;
+      x=x*x;
+      n>>=1;
+    }
+    return result;
+  }
+
+  // compute prod(1+a^(2^i), i=0..l-1)
+  TRNG_CUDA_ENABLE
+  inline lcg64::result_type lcg64::g(unsigned int l, lcg64::result_type a) {
+    lcg64::result_type p=a, res=1;
+    for (unsigned i=0; i<l; ++i) {
+      res*=1+p;
+      p*=p;
+    }
+#if ULONG_LONG_MAX>18446744073709551615ull
+    return res & 0xffffffffffffffffull;
+#else
+    return res;
+#endif
+  }
+  
+  // compute sum(a^i, i=0..s-1)
+  TRNG_CUDA_ENABLE
+  inline lcg64::result_type lcg64::f(lcg64::result_type s, lcg64::result_type a) {
+    if (s==0)
+      return 0;
+    unsigned int e=log2_floor(s);
+    lcg64::result_type y=0, p=a;
+    for (unsigned int l=0; l<=e; ++l) {
+      if (((1ull<<l) & s)>0) {
+        y=g(l, a)+p*y;
+#if ULONG_LONG_MAX>18446744073709551615ull
+	y&=0xffffffffffffffffull;
+#endif
+      }
+      p*=p;
+#if ULONG_LONG_MAX>18446744073709551615ull
+      p&=0xffffffffffffffffull;
+#endif
+    }
+    return y;
+  }
+
+  // Parallel random number generator concept
+
+  TRNG_CUDA_ENABLE
+  inline void lcg64::split(unsigned int s, unsigned int n) {
+#if !(defined __CUDA_ARCH__)
+    if (s<1 or n>=s)
+      utility::throw_this(std::invalid_argument("invalid argument for trng::lcg64::split"));
+#endif
+    if (s>1) {
+      jump(n+1);
+      P.b*=f(s, P.a);
+#if ULONG_LONG_MAX>18446744073709551615ull
+      P.b&=0xffffffffffffffffull;
+#endif
+      P.a=pow(P.a, s);
+      backward();
+    }
+  }
+
+  TRNG_CUDA_ENABLE
+  inline void lcg64::jump2(unsigned int s) {
+    S.r=S.r*pow(P.a, 1ull<<s)+f(1ull<<s, P.a)*P.b;
+#if ULONG_LONG_MAX>18446744073709551615ull
+    S.r&=0xffffffffffffffffull;
+#endif
+  }
+
+  TRNG_CUDA_ENABLE
+  inline void lcg64::jump(unsigned long long s) {
+    if (s<16) {
+      for (unsigned int i(0); i<s; ++i) 
+	step();
+    } else {
+      unsigned int i(0);
+      while (s>0) {
+	if (s%2==1)
+	  jump2(i);
+	++i;
+	s>>=1;
+      }
+    }
+  }
+
+  TRNG_CUDA_ENABLE
+  inline void lcg64::backward() {
+    for (unsigned int i(0); i<64; ++i)
+      jump2(i);
+  }
+
 }
 
 #endif
