@@ -104,60 +104,6 @@ namespace trng {
   bool operator!=(const mrg2 &R1, const mrg2 &R2) {
     return not (R1==R2);
   }
-
-  // Parallel random number generator concept
-  void mrg2::split(unsigned int s, unsigned int n) {
-    if (s<1 or n>=s)
-      throw std::invalid_argument("invalid argument for trng::mrg2::split");
-    if (s>1) {
-      jump(n+1);  long q0=S.r1;
-      jump(s);    long q1=S.r1;
-      jump(s);    long q2=S.r1;
-      jump(s);    long q3=S.r1;
-      std::vector<long> a(2), b(4);
-      a[0]=q2;  b[0]=q1;  b[1]=q0;
-      a[1]=q3;  b[2]=q2;  b[3]=q1;
-      utility::gauss(b, a, modulus);
-      P.a1=a[0];  P.a2=a[1];
-      S.r1=q1;    S.r2=q0;
-      backward();
-      backward();
-    }
-  }
-
-  void mrg2::jump2(unsigned int s) {
-    std::vector<long> b(4), c(4), d(2), r(2);
-    long t1(P.a1), t2(P.a2);
-    b[0]=P.a1;  b[1]=P.a2;
-    b[2]=1l;    b[3]=0l;
-    for (unsigned int i(0); i<s; ++i)
-      if ((i&1)==0)
-	utility::matrix_mult(b, b, c, modulus);
-      else
-	utility::matrix_mult(c, c, b, modulus);
-    r[0]=S.r1;  r[1]=S.r2;
-    if ((s&1)==0)
-      utility::matrix_vec_mult(b, r, d, modulus);
-    else
-      utility::matrix_vec_mult(c, r, d, modulus);
-    S.r1=d[0];  S.r2=d[1];
-    P.a1=t1;    P.a2=t2;
-  }
-
-  void mrg2::jump(unsigned long long s) {
-    if (s<16) {
-      for (unsigned int i(0); i<s; ++i) 
-	step();
-    } else {
-      unsigned int i(0);
-      while (s>0) {
-	if (s%2==1)
-	  jump2(i);
-	++i;
-	s>>=1;
-      }
-    }
-  }
   
   // other useful methods
   const char * const mrg2::name_str="mrg2";
@@ -165,26 +111,5 @@ namespace trng {
   const char * mrg2::name() {
     return name_str;
   }
-
-  void mrg2::backward() {
-    long t;
-    if (P.a2!=0l) {
-      t=S.r1;
-      t-=static_cast<long>((static_cast<long long>(P.a1)*
-			    static_cast<long long>(S.r2))%modulus);
-      if (t<0l)
-	t+=modulus;
-      t=static_cast<long>((static_cast<long long>(t)*
-			   static_cast<long long>
-			   (utility::modulo_invers(P.a2, modulus)))%modulus);
-    } else if (P.a1!=0l) {
-      t=S.r2;
-      t=static_cast<long>((static_cast<long long>(t)*
-			   static_cast<long long>
-			   (utility::modulo_invers(P.a1, modulus)))%modulus);
-    } else
-      t=0l;
-    S.r1=S.r2;  S.r2=t;
-  }
   
 }
